-- |
-- Module      : Crypto.Error
-- License     : BSD-style
-- Maintainer  : Vincent Hanquez <vincent@snarc.org>
-- Stability   : Stable
-- Portability : Excellent
--
module Crypto.Error
    ( module Crypto.Error.Types
    ) where

import Crypto.Error.Types
